data = ['You took too long', 'You are not easy to deal with', 'Payment Failure/Incorrect Payment',
        'You provided me with incorrect information', 'Unhappy with delay', 'Unsuitable advice',
        'You never answered my question', 'You did not understand my needs', 'I have been mis-sold',
        'My details are not accurate', 'You have asked for too much information', 'You were not helpful',
        'Payment not generated/received by customer', 'You did not keep me updated', 'Incorrect information given',
        'The performance of my product was poor', 'No reply to customer contact', 'Requested documentation not issued',
        'You did not explain the terms & conditions', 'Policy amendments not carried out',
        'You did not explain the next steps/process to me', 'I cannot understand your letter/comms',
        'Standard letter inappropriate', 'Customer payment processed incorrectly', 'All points not addressed',
        'Could not understand the agent', 'Issue with terms and conditions', 'Misleading information',
        'I can not use the customer portal', 'your customer portal is unhelpful']
